
Konzepte der Programmierung                       Universität Konstanz
Programmierkurs 2                Informatik & Informationswissenschaft
Sommer 2016                                Marc Scholl, Stefan Klinger

======================================================================

                             17. Mai 2016

----------------------------------------------------------------------
Anmerkung zum Extensionalitätsprinzip:


> f :: Int -> Int
> f a = (a+1)*2

> g :: Int -> Int
> g b = 2*b+2

Ausprobieren:

    f == g
 
Offenbar kann der Haskell-Interpreter Funktionen nicht vergleichen.

Wie steht's mit dem (um Konstanten erweiterten) λ-Kalkül?
 
    “λa. (a+1)*2”   ≠   “λb. 2*b+2”

  * Offenbar haben beide Ausdrücke unterschiedliche Struktur.

  * Der Interpreter müsste in der Lage sein allgemein abzuleiten wann
    zwei Ausdrücke “gleich” sind.


Also sind `f` und `g` verschieden?
 
  * Innendrin: Ja

  * Aber: Sie verhalten sich gleich!

 
 
Definitionen
 
    “Extensional gleich” := verhält sich gleich, hat die gleichen
                            Eigenschaften.

    “Intensional gleich” := ist gleich aufgebaut, hat die gleiche
                            Struktur.

Detaillierter (und genauer) werden diese beiden Begriffe in der Logik
definiert.


         
Wir hatten Äquivalenz von Funktionen `f`, `g` extensional definiert:

    f ≡ g   :⇔   ∀x. f x ≡ g x

Unser Beispiel:
 
      f x
    =                           Definition von `f`
      (λa. (a + 1) * 2) x
    →                           β-Reduktion
      ((a + 1) * 2)[a←x]
    ≡                           Ersetzung [a←x]
      (x + 1) * 2
    ≡                           Äquivalenzumformung
      2 * x + 2                 (übrigens: extensionale Äquivalenz)



      g x
    =                           Definition von `g`
      (λb. 2 * b + 2) x
    →                           β-Reduktion
      (2 * b + 2)[b←x]
    ≡                           Ersetzung [b←x]
      2 * x + 2



In der Informatik ist man auch an den “inneren Werten” interessiert:

    ∀x. fib x = fib' x

Meist möchte man extensionale Gleichheit behalten, aber effizienter
implementieren. ⇒ Intensional *nicht* äquivalent.



Die Abgrenzung was zum “Verhalten” gehört, ist nicht immer ganz
einfach.  Beispiel:


    Programm A: Eingabe n

        for i = 1 .. n:
            memory[i] := mod(i,256)
 
        for i = 1 .. n:
            if memory[i] != mod(i,256)
                exit("Speicher fehlerhaft")
 
        exit("Speicher OK")


    Programm B: Eingabe n

        exit("Speicher OK")

 
Frage:  Sind die Programme extensional gleich?

  * Ja, unter der Annahme dass die Auswertemaschine (insbesondere der
    Speicher) korrekt funktioniert.

  * Nein, denn tatsächlich gehört bei beiden Beispielen oben der
    Gesundheitszustand des Speichers mit zur Eingabe.

  * Wir gehen immer davon aus dass die Maschine korrekt funktioniert.


- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Anwendung im λ-Kalkül


Erinnerung:

    f ≡ g   :⇔   ∀x. f x ≡ g x


η-Konversion: Sei `e` eine Funktion, Variable `x` nicht frei in `e`.

    e  ↔  λx. e x    — η-Konversion



Haben damit:

    α-Konversion

        λx. e   ↔   λy. e[x←y]      wenn `y` nicht frei in `e` ist

    β-Reduktion

        (λx. e1) e2   →   e1[x←e2]

    δ-Reduktion

        + e   →   ‹+› e             ‹+› ist die Maschinenoperation

    η-Konversion

        e ↔ λx. e x                 wenn `e` Funktion, und
                                    `x` nicht frei in `e` ist



Beispiele:

> foo x = div 42 x

      foo
    =
      λx. div 42 x
    ≡                           Applikation ist linksassoziativ
      λx. (div 42) x
    ↔                           η-Konversion, `x` ∉ free `div 42`
      div 42

> foo' = div 42



Sei `fact` nochmal die Fakultät.
 
> fact :: Integer -> Integer
> fact 0 = 1
> fact n = n * fact (n-1)

Betrachten:
 
> bar n = div $ fact n

Was tut die Funktion?

                     n!
    bar n   ≡   λx. ---
                     x

Kann man das `n` weg-η-konvertieren?

      bar
    =
      λn. div $ fact n
    ↔                      η-Konversion ????
      div $ fact

Nein, das ist Mumpitz!
 
      bar
    =
      λn. div $ fact n
    =
      λn. div (fact n)

Das hat die Struktur

                 λn
                 |
    bar  =       @
                / \
             div   @
                  / \
              fact   n    <- hier ist das Argument `n`

Vergleiche mit `foo x = div 42 x`:

                 λx
                 |
    foo  =       @
                / \
               @   x  <- hier ist das Argument `x`
              / \
           div   42
 

Für η-Konversion brauchen wir die Struktur

           λx
           |
           @             also    λx. e x    wo  x ∉ free e
          / \
      hier    x
     kommt
    x nicht
    frei vor

 

Manchmal ist es möglich, aber nicht ganz einfach, eine Variable in die
Position “rechts oben” zu verschieben:

> nextfib :: (Integer,Integer) -> (Integer,Integer)
> nextfib (x,y) = (y,x+y)

> fib'4 n = fst . head . drop n $ iterate nextfib (0,1)

Das `n` könnte man loswerden wollen.



    flip :: (a -> b -> c) -> b -> a -> c

Was tut `flip`?  Wie könnte es definiert sein?

    flip f x y = f y x
 

Funktion `flip f` ist wie `f` mit vertauschten Argumenten!

Ausprobieren:

    flip ($) 9 sqrt


Damit können wir `n` nach “rechts oben” schieben:

      fib'4
    =
      \n -> fst . head . drop n $ iterate nextfib (0,1)
    =
      \n -> fst ( head ( drop n (iterate nextfib (0,1)) ))
    =
      \n -> fst ( head ( flip drop (iterate nextfib (0,1)) n ))
    =                    .neue Klammern                    .
      \n -> fst ( head ( (flip drop (iterate nextfib (0,1))) n ))
    =
      \n -> fst . head . flip drop (iterate nextfib (0,1)) $ n

    ↔    | η-Konversion,
         | `n` ∉ free `fst . head . flip drop (iterate nextfib (0,1))`
 
      fst . head . flip drop (iterate nextfib (0,1))
 
> fib'5 = fst . head . flip drop (iterate nextfib (0,1))

Man nennt das “point-free style” weil keine λ-Abstraktionen mehr
vorkommen.

Später: Man kann λ-Abstraktionen ganz aus Programmen entfernen.
        Interessant für Kompilerbau.


----------------------------------------------------------------------
Zu Aufgabe 1, crossSum


Tutorien:

> crossSum :: Integer -> Integer
> crossSum 0 = 0
> crossSum n = let (d,m) = divMod n 10 in m + crossSum d


- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -


Erweitern wir die Aufgabe für ganze Zahlen ℤ: Was ist mit negativen
Zahlen?

    -123 = -1 * 100 + (-2) * 10 + (-3) * 1

    crossSum (-123)  →  -6   ???

Wieso nicht?

    div (-123) 10

Stimmt das?

Die Funktionen `rem` und `quot` sind eng mit den Funktionen `mod` und
`div` verwandt. Es gelten nämlich die beiden Gelichungen

  (x `quot` y)*y + (x `rem` y) == x
  (x `div`  y)*y + (x `mod` y) == x

Allerdings runded `quot` immer in Richtung 0 während `div` immer in
Richtung -∞ rundet. D.h.,

     div 15 6
  ==                                       div  (-15) 6 == ???
     2                   aber           !=
  ==                                       quot (-15) 6 == ???
     quot 15 6



                 -15                            15
   -∞ ---.-----.--'--.-----.-----|-----.-----.--'--.-----.--> ∞
        -24   -18   -12   -6     0     6     12    18    24
               |     |                       |
            div| ←'  |                    div| ←'
                     |                   quot| ←'
                  '→ |quot


Insbesondere ist also

  mod  (-15) 6 == 2         und         rem  (-15) 6 == -3


Für `quot`, `rem` gibt es, ebenso wie für `div`, `mod` eine
zusammenfassende Funktionen `quotRem`, für die gilt:

  (q,r) == quotRem x y  ⇔  q = quot x y  ∧  r == rem x y

> crossSum' :: Integer -> Integer
> crossSum' 0 = 0
> crossSum' n = let (q,r) = quotRem n 10 in r + crossSum' q


- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -


Natürlich kann man das Problem noch weiter zerlegen:

> digits' :: Integer -> Integer -> [Integer]

> digits' _ 0 = []
> digits' b n = let (q,r) = quotRem n b
>               in r : digits' b q

Frage: Was tut diese Funktion ???

    digits' 10 1234
    digits' 10 (-1234)

    digits' 2 10
    digits' 2 255

Die Quersumme ist dann einfach

> crossSum'' = sum . digits' 10


Für die Berechnung der Quersumme ist die Reihenfolge der Ziffern
irrelevant.  Trozdem:

Wie können wir die Zahlen in der richtigen Reihenfolge ausgeben?

    digits'' 10 1234
  → 
    [1,2,3,4]


Plan A: `reverse`
-----------------

> digits'' :: Integer -> Integer -> [Integer]

 > digits'' b n = reverse (digits' b n)

Wie kann man das umformulieren mit `$` und `.`?

  digits'' b n = reverse $ digits' b n

  digits'' b n = reverse $ (digits' b) n

  digits'' b n = (reverse . digits' b) n

  digits'' b = \n-> (reverse . digits' b) n

 > digits'' b = reverse . digits' b

Bis hier hin haben wir den Term durch Umformulieren vereinfacht.  Man
kann das aber auch übertreiben:

  digits'' b = (.) reverse (digits' b)

  digits'' b = ((.) reverse) (digits' b)

  digits'' b = ((.) reverse) . digits') b

> digits'' = (.) reverse . digits'


Testen mit

> pack :: Integer -> [Integer] -> Integer
> pack b = foldl (\n d -> n * b + d) 0

    pack 10 $ digits'' 10 1234

Wie funktioniert `pack`?



Plan B:  Akkumulierender Parameter
----------------------------------

> digitsAcc :: Integer -> [Integer] -> Integer -> [Integer]

> digitsAcc _ rs 0 = rs
> digitsAcc b rs n = let (q,r) = quotRem n b
>                    in digitsAcc b (r:rs) q

Wie kann man das aufrufen um die Dezimalstellen der Zahl 1234 zu
bekommen?

    digitsAcc 10 [] 1234

Wir verstecken `digitsAcc` besser in einer lokalen Definition, und
nennen es `go`.

> digits''' :: Integer -> Integer -> [Integer]

 > digits''' b x = go b [] x
 >   where
 >     go _ rs 0 = rs
 >     go b rs n = let (q,r) = quotRem n b
 >                 in go b (r:rs) q

    digits''' 10 1234

Im der Rekursion von `go` wird `b` immer den gleichen Wert haben.  Wir
m üssen es also nicht immer explizit weiterreichen, sondern greifen
auf das Argument von `digits'''` zu:

 > digits''' b x = go [] x
 >   where
 >     go rs 0 = rs
 >     go rs n = let (q,r) = quotRem n b
 >               in go (r:rs) q

Und in der esten Zeile wird man das `x` durch η-Konversion los:

> digits''' b = go []
>   where
>     go rs 0 = rs
>     go rs n = let (q,r) = quotRem n b
>               in go (r:rs) q

Kleine Unschönheiten wie

    digits''' 1 123

    digits''' 10 0

können durch eine “Wrapper”-Funktion beseitigt werden:

Die Funktion `digits b` berechnet die Darstellung einer gegebenen
Zahl zur Basis `b`, angefangen mit der niederwertigsten Stelle. Für
die 0 wird die Liste [0] ausgegeben, eine Basis b<2 erzeugt einen
Fehler.

> digits :: Integer -> Integer -> [Integer]

> digits b n

      Wenn die Basis zu klein ist brechen wir die Berechnung ab.

>     | b < 2 = error "base to small"

      Die Null stellen wir als einfache Null dar

>     | n == 0 = [0]

      Und sonst rufen wir die eigentliche Arbeits-Funktion auf:

>     | otherwise = digits''' b n

Wrapper sind immer dann sinnvoll, wenn Aufrufparameter einmalig
getestet werden müssen, statt bei jedem Rekursionsaufruf.


----------------------------------------------------------------------
Darstellung von Zahlen im einfachen λ-Kalkül


Bisher gesehen:

  * Paare im λ-Kalkül (Vorlesung).

  * Boolesche Algebra im λ-Kalkül (frühere Übung).



Jetzt:  Definition von Zahlen.

    0 = λf x. x

    1 = λf x. f x
 
    2 = λf x. f (f x)

    3 = λf x. f (f (f x))

    ...

Was ist die Idee hier?

    * Eine Zahl ist eine Funktion (was anderes kennen wir ja nicht).

    * Zahl `n` wendet ihr Argument `f` n-mal an.

Mit Komposition geschrieben:

    0 = λf. id

    1 = λf. f
 
    2 = λf. f ∘ f

    3 = λf. f ∘ f ∘ f



Wie sieht die Nachfolgerfunktion (+ 1) aus?

    suc = λn. ???

    suc = λn. λf. f ∘ n f


Ausprobieren:

> nul = \f -> id
 
> suc n = \f -> f . n f

    nul (+1) 0

    (suc nul) (+1) 0

> encode :: Integer -> (t -> t) -> t -> t
> encode 0 = nul
> encode n = suc $ encode (n-1)

> decode :: ((Integer -> Integer) -> Integer -> Integer) -> Integer
> decode n = n (+1) 0

    decode $ encode 42



Wie sieht die Addition aus?

    add = λm n. λf. m f ∘ n f


Was ist die Idee dabei?

    `f` erst n-mal anwenden, dann noch m-mal.

      add 2 3

    →     mit Definition von `add` und Komposition

      (λn m f x. n f (m f x)) 2 3
    →
      (λm f x. 2 f (m f x)) 3
    →
      λf x. 2 f (3 f x)

    =     Definition von `2`
            ...............
      λf x. (λf x. f (f x)) f (3 f x)
    →
      λf x. (λx. f (f x)) (3 f x)
    →
      λf x. f (f (3 f x))

    =     Definition von `3`
                  ...................
      λf x. f (f ((λf x. f (f (f x))) f x))
    →
      λf x. f (f (f (f (f x))))


Addition von oben:

    add = λm n. λf. m f ∘ n f

> add m n = \f -> m f . n f

    decode $ encode 23 `add` encode 42


Schön.  Schön schön.  Multiplikation?

    mul = λm n. λf. m (n f)

Was ist hier die Idee?

  * `n f` heißt `f` n-mal anwenden.

  * `m (n f)` heißt `n f` m-mal anwenden, also `f` m*n-mal anwenden.

> mul m n = \f -> m (n f)

    decode $ encode 6 `mul` encode 7

Man hätte auch schreiben können:  mul' m n = m . n

    decode $ encode 6 . encode 7

Potenzieren ist auch noch einfach:

    pwr = λm n.  ???

> pwr m n = n m

    decode $ pwr (encode 10) (encode 3)


Division?  Ziemlich schwierig.

    div = λn. (λf. (λx. x x) (λx. f (x x))) (λc n m f x. (λd. (λn. n
          (λx a b. b) (λa b. a)) d ((λf x. x) f x) (f (c d m f x)))
          ((λm n. n (λn f x. n (λg h. h (g f)) (λu. x) (λu. u)) m) n m))
          ((λn f x. f (n f x)) n)

Eine Herleitung gibt's z.B. bei Wikipedia [1], diese funktioniert
allerdings nicht direkt mit Haskell (wegen Rekursion und Typsystem;
cf. später).



=> Auch “Zahlen” lassen sich selbst als Programm darstellen!


  * Das macht natürlich niemand, weil unsere Rechner-Architektur
    Primitive für die Arithmetik bereitstellt.

  * Prinzipiell ist der Unterschied zwischen Daten und Kontrollfluss
    aber fließend!

  * Man kann auch im einfachen λ-Kalkül rechnen.  Das gezeigte
    Encoding ist als “Church-Encoding” bekannt [1,2].


----------------------------------------------------------------------

[1] Church encoding
    https://en.wikipedia.org/wiki/Church_encoding

[2] Church-Kodierung
    https://de.wikipedia.org/wiki/Church-Kodierung

======================================================================
